section \<open>Simple Memory Model\<close>
theory Simple_Memory
imports "../../lib/LLVM_Integer" "../../lib/LLVM_Double"  "../../lib/MM/MMonad" 
begin

  text \<open>Here, we combine a model of LLVM values, with our generic block-based memory model\<close>

  datatype llvm_ptr = is_null: PTR_NULL | is_addr: PTR_ADDR (the_addr: addr)
  hide_const (open) llvm_ptr.is_null llvm_ptr.is_addr llvm_ptr.the_addr

    
  lifting_update memory.lifting
  lifting_forget memory.lifting
        
  subsection \<open>LLVM Values\<close>

  datatype llvm_val = 
    is_struct: LL_STRUCT (the_fields: "llvm_val list") 
  | is_int: LL_INT (the_int: lint) 
  | is_double: LL_DOUBLE (the_double: double) (* 
      TODO: Similar to lint, we could encode different floating-point layouts here, 
        and restrict the code-generator to only accept the ones supported by LLVM.
    *)
  | is_ptr: LL_PTR (the_ptr: llvm_ptr)
  hide_const (open) 
    llvm_val.is_struct llvm_val.the_fields
    llvm_val.is_int llvm_val.the_int
    llvm_val.is_double llvm_val.the_double
    llvm_val.is_ptr llvm_val.the_ptr
  
  
  datatype llvm_struct = 
    is_struct: VS_STRUCT (the_fields: "llvm_struct list") 
  | is_int: VS_INT (the_width: nat)
  | is_double: VS_DOUBLE
  | is_ptr: VS_PTR 
  hide_const (open) 
    llvm_struct.is_struct llvm_struct.the_fields
    llvm_struct.is_ptr
    llvm_struct.is_double
    llvm_struct.is_int llvm_struct.the_width
  
  
  
  fun llvm_struct_of_val where
    "llvm_struct_of_val (LL_STRUCT vs) = VS_STRUCT (map llvm_struct_of_val vs)"
  | "llvm_struct_of_val (LL_INT i) = VS_INT (width i)"
  | "llvm_struct_of_val (LL_DOUBLE _) = VS_DOUBLE"
  | "llvm_struct_of_val (LL_PTR _) = VS_PTR"

  fun llvm_zero_initializer where
    "llvm_zero_initializer (VS_STRUCT vss) = LL_STRUCT (map llvm_zero_initializer vss)"
  | "llvm_zero_initializer (VS_INT w) = LL_INT (lconst w 0)"
  | "llvm_zero_initializer (VS_DOUBLE) = LL_DOUBLE (double_of_word 0)"
  | "llvm_zero_initializer VS_PTR = LL_PTR PTR_NULL"
  
  lemma struct_of_llvm_zero_initializer[simp]: "llvm_struct_of_val (llvm_zero_initializer s) = s"
    apply (induction s) 
    apply (simp_all add: map_idI)
    done

  (*type_synonym llvm_memory = "llvm_val memory"
  translations (type) "llvm_memory" \<leftharpoondown> (type) "llvm_val memory"
  *)

  type_synonym 'a llM = "('a,llvm_val) M"
  translations
    (type) "'a llM" \<leftharpoondown> (type) "('a, llvm_val) M"

    
  subsection \<open>Raw operations on values\<close>  
  context
    includes monad_syntax_M
  begin
    
  definition llvm_extract_addr :: "llvm_val \<Rightarrow> addr llM" where
    "llvm_extract_addr v \<equiv> case v of LL_PTR (PTR_ADDR a) \<Rightarrow> return a | _ \<Rightarrow> fail"

  definition llvm_extract_ptr :: "llvm_val \<Rightarrow> llvm_ptr llM" where
    "llvm_extract_ptr v \<equiv> case v of LL_PTR p \<Rightarrow> return p | _ \<Rightarrow> fail"
    
  definition llvm_extract_sint :: "llvm_val \<Rightarrow> int llM" where
    "llvm_extract_sint v \<equiv> case v of LL_INT i \<Rightarrow> return (lint_to_sint i) | _ \<Rightarrow> fail" 
        
  definition llvm_extract_unat :: "llvm_val \<Rightarrow> nat llM" where
    "llvm_extract_unat v \<equiv> case v of LL_INT i \<Rightarrow> return (nat (lint_to_uint i)) | _ \<Rightarrow> fail" 

  definition llvm_extract_value :: "llvm_val \<Rightarrow> nat \<Rightarrow> llvm_val llM" where 
  "llvm_extract_value v i \<equiv> case v of 
    LL_STRUCT vs \<Rightarrow> doM {
      assert (i<length vs);
      return (vs!i)
    }
  | _ \<Rightarrow> fail"
      
  definition llvm_insert_value :: "llvm_val \<Rightarrow> llvm_val \<Rightarrow> nat \<Rightarrow> llvm_val llM" where 
  "llvm_insert_value v x i \<equiv> case v of 
    LL_STRUCT vs \<Rightarrow> doM {
      assert (i<length vs);
      assert (llvm_struct_of_val (vs!i) = llvm_struct_of_val x);
      return (LL_STRUCT (vs[i:=x]))
    }
  | _ \<Rightarrow> fail"

    
  subsection \<open>Interface functions\<close>
  
  subsubsection \<open>Typed arguments\<close>
    
  (* TODO: redundancy with is_valid_addr! *)
  definition llvmt_check_addr :: "addr \<Rightarrow> unit llM" where "llvmt_check_addr a \<equiv> doM { 
    Mvalid_addr a
  }"
    
  definition llvmt_load :: "addr \<Rightarrow> llvm_val llM" where "llvmt_load a \<equiv> doM { 
    Mload a
  }"
  
  definition "llvmt_store x a \<equiv> doM { 
    xorig \<leftarrow> llvmt_load a; 
    assert llvm_struct_of_val x = llvm_struct_of_val xorig;
    Mstore a x
  }"
  
  definition "llvmt_alloc s n \<equiv> doM {
    Mmalloc (replicate n (llvm_zero_initializer s))
  }"
  
  definition llvmt_free :: "nat \<Rightarrow> unit llM" where "llvmt_free b \<equiv> doM {
    Mfree b
  }"  

  definition "llvmt_freep p \<equiv> doM {
    assert llvm_ptr.is_addr p;
    let a = llvm_ptr.the_addr p;
  
    assert addr.index a=0;
    llvmt_free (addr.block a);
    return ()
  }"  

  definition "llvmt_allocp s n \<equiv> doM {
    b \<leftarrow> llvmt_alloc s n;
    return (PTR_ADDR (ADDR b 0))
  }"
    
  
  definition llvmt_check_ptr :: "llvm_ptr \<Rightarrow> unit llM" where "llvmt_check_ptr p \<equiv> 
    if llvm_ptr.is_null p then return ()
    else doM {
      let a = llvm_ptr.the_addr p;
      Mvalid_addr a \<comment> \<open>TODO: support 1-beyond-end pointers!\<close>
    }"
      
  definition "llvmt_ofs_ptr p ofs \<equiv> doM {
    assert (llvm_ptr.is_addr p);
    let a = llvm_ptr.the_addr p;
    let b = addr.block a;
    let i = addr.index a;
    let i = i + ofs;
    let r = PTR_ADDR (ADDR b i);
    llvmt_check_ptr r;
    return r
  }"  
    
  definition "llvmt_check_ptrcmp p\<^sub>1 p\<^sub>2 \<equiv> 
    if p\<^sub>1=PTR_NULL \<or> p\<^sub>2=PTR_NULL then 
      return () 
    else doM {
      llvmt_check_ptr p\<^sub>1;
      llvmt_check_ptr p\<^sub>2
    }"
  
  definition "llvmt_ptr_eq p\<^sub>1 p\<^sub>2 \<equiv> doM {
    llvmt_check_ptrcmp p\<^sub>1 p\<^sub>2;
    return (p\<^sub>1 = p\<^sub>2)
  }"
  
  definition "llvmt_ptr_neq p\<^sub>1 p\<^sub>2 \<equiv> doM {
    llvmt_check_ptrcmp p\<^sub>1 p\<^sub>2;
    return (p\<^sub>1 \<noteq> p\<^sub>2)
  }"
  
  subsubsection \<open>Embedded arguments\<close>

  definition "llvm_load a \<equiv> doM {
    a \<leftarrow> llvm_extract_addr a;
    llvmt_load a
  }"
  
  definition "llvm_store x a \<equiv> doM {
    a \<leftarrow> llvm_extract_addr a;
    llvmt_store x a
  }"
  
  definition "llvm_alloc s n \<equiv> doM {
    n \<leftarrow> llvm_extract_unat n;
    p \<leftarrow> llvmt_allocp s n;
    return (LL_PTR p)
  }"

  definition "llvm_extract_base_block a \<equiv> case a of ADDR b i \<Rightarrow> if i=0 then return b else fail"
  
  definition "llvm_free p \<equiv> doM {
    p \<leftarrow> llvm_extract_ptr p;
    llvmt_freep p
  }"
  
  definition "llvm_ofs_ptr p ofs \<equiv> doM {
    p \<leftarrow> llvm_extract_ptr p;
    ofs \<leftarrow> llvm_extract_sint ofs;
    r \<leftarrow> llvmt_ofs_ptr p ofs;
    return (LL_PTR r)
  }"  
  
  
  definition "llvm_ptr_eq p\<^sub>1 p\<^sub>2 \<equiv> doM {
    p\<^sub>1 \<leftarrow> llvm_extract_ptr p\<^sub>1;
    p\<^sub>2 \<leftarrow> llvm_extract_ptr p\<^sub>2;
    r \<leftarrow> llvmt_ptr_eq p\<^sub>1 p\<^sub>2;
    return (LL_INT (bool_to_lint r))
  }"  
  
  definition "llvm_ptr_neq p\<^sub>1 p\<^sub>2 \<equiv> doM {
    p\<^sub>1 \<leftarrow> llvm_extract_ptr p\<^sub>1;
    p\<^sub>2 \<leftarrow> llvm_extract_ptr p\<^sub>2;
    r \<leftarrow> llvmt_ptr_neq p\<^sub>1 p\<^sub>2;
    return (LL_INT (bool_to_lint r))
  }"  
  
end
end
