theory Generic_Memory
imports NEMonad
begin

  datatype 'v cell = FRESH | FREED | is_block: BLOCK (vals: "'v list")
  hide_const (open) is_block vals

  datatype addr = ADDR (block: nat) (index: nat) 
  hide_const (open) block index
  
  class infinite =
    assumes infinite_UNIV[simp,intro!]: "\<not> finite (UNIV :: 'a set)"

  instance nat :: infinite  
    apply standard
    by simp
      
  type_synonym block = nat  
    
  typedef 'v memory = "{ f :: block \<Rightarrow> 'v cell . finite {a. f a \<noteq> FRESH} }" 
    by (auto intro: exI[where x="\<lambda>_. FRESH"])

  setup_lifting type_definition_memory  

  lift_definition mapp :: "'v memory \<Rightarrow> block \<Rightarrow> 'v cell" is "\<lambda>m a. m a" .
  lift_definition mmap :: "'v memory \<Rightarrow> block \<Rightarrow> ('v cell \<Rightarrow> 'v cell) \<Rightarrow> 'v memory" is "\<lambda>m a f. m(a := f (m a))"
  proof goal_cases
    case (1 f a fx)
    thm 1
    have "{aa. (f(a := cell)) aa \<noteq> FRESH} \<subseteq> insert a {a. f a \<noteq> FRESH}" for cell
      by auto
    moreover from 1 have "finite \<dots>" by auto
    ultimately show ?case by (meson finite_subset)
  qed
    
  lemma mapp_mmap[simp]: "mapp (mmap f a g) = (mapp f)(a:=g (mapp f a))"
    apply transfer
    by auto

  lemma mext: "f=f' \<longleftrightarrow> (\<forall>a. mapp f a = mapp f' a)"
    apply transfer
    by auto

  definition "block_size m b \<equiv> case mapp m b of 
    BLOCK vs \<Rightarrow> length vs
  | _ \<Rightarrow> 0"
    
  datatype block_state = FRESH | ALLOC | FREED
  hide_const (open) FRESH ALLOC FREED
    
  definition "block_state m b \<equiv> case mapp m b of 
    FRESH \<Rightarrow> block_state.FRESH
  | BLOCK _ \<Rightarrow> block_state.ALLOC
  | FREED \<Rightarrow> block_state.FREED"
  
  abbreviation "is_FRESH m b \<equiv> block_state m b = block_state.FRESH"  
  abbreviation "is_ALLOC m b \<equiv> block_state m b = block_state.ALLOC"  
  abbreviation "is_FREED m b \<equiv> block_state m b = block_state.FREED"  
  
        
  definition "get_addr m a \<equiv> 
    case a of ADDR b i \<Rightarrow> cell.vals (mapp m b) ! i
  "  
    
  text \<open>The \<open>put\<close> function is closed to do nothing on invalid addresses. 
    This removes a lot of side-conditions from the lens laws below, 
    making them simpler to reason with.\<close>
  definition "put_addr m a x \<equiv> 
    case a of ADDR b i \<Rightarrow> mmap m b (\<lambda>BLOCK vs \<Rightarrow> BLOCK (vs[i:=x]) | c \<Rightarrow> c )
  "

  lemma block_state_put[simp]:
    "block_state (put_addr m a x) b = block_state m b"
    "block_size (put_addr m a x) b = block_size m b"
    unfolding block_state_def block_size_def put_addr_def 
    by (auto split: addr.splits cell.splits simp: mext)
  
  lemma addr_put_get[simp]: "put_addr m a (get_addr m a) = m"
    and addr_put_put[simp]: "put_addr (put_addr m a y) a x = put_addr m a x"
    unfolding get_addr_def put_addr_def 
    by (auto split: addr.splits cell.splits simp: mext)
    
    
  definition "is_valid_addr m a \<equiv> 
    case a of ADDR b i \<Rightarrow> i<block_size m b"  
      
  lemma addr_get_put[simp]: "is_valid_addr m a \<Longrightarrow> get_addr (put_addr m a x) a = x"
    and addr_put_valid[simp]: "is_valid_addr m a \<Longrightarrow> is_valid_addr (put_addr m a x) a"
    unfolding is_valid_addr_def block_size_def get_addr_def put_addr_def
    by (auto split: addr.splits cell.splits)
    
  lemma addr_indep[simp]: "\<lbrakk> a\<^sub>1\<noteq>a\<^sub>2\<rbrakk> \<Longrightarrow> get_addr (put_addr m a\<^sub>1 x) a\<^sub>2 = get_addr m a\<^sub>2"    
    unfolding is_valid_addr_def get_addr_def put_addr_def 
    by (auto split: addr.splits cell.splits simp: mext nth_list_update')
  
  lemma addr_ext_iff: "m=m' \<longleftrightarrow> (
    (\<forall>b. block_state m b = block_state m' b)
  \<and> (\<forall>b. block_size m b = block_size m' b)
  \<and> (\<forall>a. is_valid_addr m a \<and> is_valid_addr m' a \<longrightarrow> get_addr m a = get_addr m' a)  
  )"
    apply (rule; clarsimp)
    apply (clarsimp simp: mext)
    subgoal for a
      apply (cases "mapp m a"; cases "mapp m' a"; simp)
      apply (auto 
        simp: block_state_def block_size_def get_addr_def is_valid_addr_def
        split: addr.splits cell.splits
        )
      by (metis cell.sel nth_equalityI)
    done
    
  definition list_combine where
    "list_combine xs\<^sub>1 I xs\<^sub>2 \<equiv> map (\<lambda>i. if i\<in>I then xs\<^sub>2!i else xs\<^sub>1!i) [0..<length xs\<^sub>1 ]"
    
  lemma nth_undefined: "\<not>i<length xs \<Longrightarrow> xs!i = undefined (i-length xs)"
    apply (induction xs arbitrary: i)
    apply (simp add: nth_def)
    by auto
    
  lemma list_combine_nth: "length xs\<^sub>1 = length xs\<^sub>2 \<Longrightarrow> list_combine xs\<^sub>1 I xs\<^sub>2 ! i = (if i\<in>I then xs\<^sub>2!i else xs\<^sub>1!i)"  
    unfolding list_combine_def
    apply (cases "i<length xs\<^sub>2")
    subgoal by (auto simp: )
    subgoal by (simp add: nth_undefined)
    done
    
  lemma list_combine_length[simp]: "length (list_combine xs\<^sub>1 I xs\<^sub>2) = length xs\<^sub>1"  
    by (auto simp: list_combine_def)
    
    
    
  fun cell_combine where
    "cell_combine FRESH I c = c"
  | "cell_combine c I FRESH = c"
  | "cell_combine FREED I c = FREED"
  | "cell_combine c I FREED = FREED"
  | "cell_combine (BLOCK vs\<^sub>1) I (BLOCK vs\<^sub>2) = BLOCK (list_combine vs\<^sub>1 I vs\<^sub>2)"
    
  lemma cell_combine_add_simps[simp]:
    "cell_combine c I FRESH = c"
    "cell_combine FREED I c = FREED"
    "cell_combine c I FREED = FREED"
    by (cases c; simp)+
  
  lift_definition mcombine :: "('v) memory \<Rightarrow> (addr set) \<Rightarrow> ('v) memory \<Rightarrow> ('v) memory" is
    "\<lambda>m\<^sub>1 A m\<^sub>2 b. cell_combine (m\<^sub>1 b) {i. ADDR b i \<in> A} (m\<^sub>2 b)"
    subgoal for f\<^sub>1 A f\<^sub>2
    proof goal_cases
      case 1
      
      have "{a. cell_combine (f\<^sub>1 a) {i. ADDR a i \<in> A} (f\<^sub>2 a) \<noteq> FRESH} \<subseteq> {a. f\<^sub>1 a \<noteq> FRESH} \<union> {a. f\<^sub>2 a \<noteq> FRESH}"
        by auto
        
      from finite_subset[OF this] 1 show ?case by simp 
    qed
  done
    
  
  lemma mapp_combine[simp]: "mapp (mcombine m\<^sub>1 A m\<^sub>2) b = cell_combine (mapp m\<^sub>1 b) {i. ADDR b i \<in> A} (mapp m\<^sub>2 b)"
    apply transfer
    by auto  

    
  lemma block_state_combine[simp]: "block_state (mcombine m\<^sub>1 A m\<^sub>2) b = (
    case (block_state m\<^sub>1 b, block_state m\<^sub>2 b) of
      (block_state.FRESH, bs\<^sub>2) \<Rightarrow> bs\<^sub>2
    | (bs\<^sub>1, block_state.FRESH) \<Rightarrow> bs\<^sub>1
    | (block_state.FREED, _) \<Rightarrow> block_state.FREED
    | (_, block_state.FREED) \<Rightarrow> block_state.FREED
    | _ \<Rightarrow> block_state.ALLOC
  )"  
    unfolding block_state_def
    by (auto split: cell.split)
  
  lemma block_size_combine[simp]: "block_size (mcombine m\<^sub>1 A m\<^sub>2) b = (
    case (block_state m\<^sub>1 b, block_state m\<^sub>2 b) of
      (block_state.FRESH, bs\<^sub>2) \<Rightarrow> block_size m\<^sub>2 b
    | (bs\<^sub>1, block_state.FRESH) \<Rightarrow> block_size m\<^sub>1 b
    | (block_state.FREED, _) \<Rightarrow> 0
    | (_, block_state.FREED) \<Rightarrow> 0
    | _ \<Rightarrow> block_size m\<^sub>1 b
  )"  
    unfolding block_state_def block_size_def
    by (auto split: cell.split)

  lemma block_size_not_alloc[simp]: "block_state m b \<noteq> block_state.ALLOC \<Longrightarrow> block_size m b = 0"  
    unfolding block_state_def block_size_def
    by (auto split: cell.split)
    
    
  lemma finite_non_fresh: "finite {a. mapp f a \<noteq> FRESH}"  
    apply transfer
    .
    
  lemma ex_fresh: "finite A \<Longrightarrow> \<exists>a\<in>-A. mapp m a = FRESH"
    apply transfer
    by (metis (mono_tags, lifting) finite_compl infinite_UNIV rev_finite_subset subset_Collect_conv)
      

  (* Valid transitions of memory cells *)  
  definition "cell_state_trans c c' \<equiv> 
    (cell.is_block c \<longrightarrow> (cell.is_block c' \<and> length (cell.vals c') = length (cell.vals c)) \<or> c'=FREED)
  \<and> (c=FREED \<longrightarrow> c'=FREED)"

      
  (* Alternative characterization: *)  
  inductive cell_state_trans1 where
    "cell_state_trans1 FRESH (BLOCK vs)" 
  | "length vs' = length vs \<Longrightarrow> cell_state_trans1 (BLOCK vs) (BLOCK vs')" 
  | "cell_state_trans1 (BLOCK vs) FREED"
    
  (* TODO: Move *)  
  lemma r2_into_rtranclp: "P a b \<Longrightarrow> P b c \<Longrightarrow> P\<^sup>*\<^sup>* a c" by auto
    
  lemma cell_state_trans_alt: "cell_state_trans c c' = cell_state_trans1\<^sup>*\<^sup>* c c'"
  proof
    show "cell_state_trans1\<^sup>*\<^sup>* c c' \<Longrightarrow> cell_state_trans c c'"
      apply (induction rule: rtranclp.induct)
      unfolding cell_state_trans_def
      by (auto elim: cell_state_trans1.cases)
  
    show "cell_state_trans c c' \<Longrightarrow> cell_state_trans1\<^sup>*\<^sup>* c c'"  
      unfolding cell_state_trans_def
      apply (cases c; cases c'; simp)
      apply (auto intro: r2_into_rtranclp cell_state_trans1.intros)
      done
    
  qed    
  

  
  lemma cell_state_trans_simps[simp]:
    "cell_state_trans c c"
    "cell_state_trans FRESH c"
    "cell_state_trans c FREED"
    "cell_state_trans (BLOCK vs) (BLOCK vs') \<longleftrightarrow> length vs' = length vs"
    "cell_state_trans FREED c' \<longleftrightarrow> c'=FREED"
    by (auto simp: cell_state_trans_def)
  
      
  lemma cell_state_trans_trans[trans]: "cell_state_trans c c' \<Longrightarrow> cell_state_trans c' c'' \<Longrightarrow> cell_state_trans c c''"  
    by (auto simp: cell_state_trans_def)
  
  
  definition "valid_block_trans m m' b \<equiv> 
      (is_ALLOC m b \<longrightarrow> (is_ALLOC m' b \<and> block_size m' b = block_size m b) \<or> is_FREED m' b)
    \<and> (is_FREED m b \<longrightarrow> is_FREED m' b)"  
    
  lemma valid_block_trans_alt: "valid_block_trans m m' b \<longleftrightarrow> cell_state_trans (mapp m b) (mapp m' b)"  
    unfolding valid_block_trans_def cell_state_trans_def block_state_def block_size_def
    by (auto split: cell.split)
    

  lemma valid_block_trans_refl[simp]:
    "valid_block_trans m m b"
    by (auto simp: valid_block_trans_def)
  
      
  lemma valid_block_trans_trans[trans]: "valid_block_trans c c' b \<Longrightarrow> valid_block_trans c' c'' b \<Longrightarrow> valid_block_trans c c'' b"  
    by (auto simp: valid_block_trans_def)
    
    
      
  definition "alloc_blocks m m' \<equiv> {b . is_FRESH m b \<and> \<not>is_FRESH m' b}"

  definition "freed_blocks m m' \<equiv> {b . \<not>is_FREED m b \<and> is_FREED m' b}"
  
    
  lemma alloc_blocks_same[simp]: "alloc_blocks m m = {}"
    by (auto simp: alloc_blocks_def)
  
  
  datatype interference = interference (r: "addr set") (w: "addr set") (e: "block set")
  hide_const (open) r w e 
    
  abbreviation "intf_r r \<equiv> interference {r} {} {}" \<comment> \<open>Read of address\<close>
  abbreviation "intf_w w \<equiv> interference {} {w} {}" \<comment> \<open>Write on address\<close>
  abbreviation "intf_e e \<equiv> interference {} {} {e}" \<comment> \<open>(read) reference to block\<close>
  
  
  instantiation interference :: comm_monoid_add
  begin
    definition "0 = interference {} {} {}"
    definition "a+b \<equiv> case (a,b) of (interference ra wa ea, interference rb wb eb) \<Rightarrow>
      interference (ra\<union>rb) (wa\<union>wb) (ea\<union>eb)"

     instance
       apply standard     
       unfolding zero_interference_def plus_interference_def
       by (auto split: interference.split)
      
  end
  
    
  lemma interference_zero_simps[simp]:  
    "interference.r 0 = {}"  
    "interference.w 0 = {}"  
    "interference.e 0 = {}"  
    by (auto split: interference.split simp: zero_interference_def)
    
  lemma interference_plus_simps[simp]:
    "interference.r (a+b) = interference.r a \<union> interference.r b"  
    "interference.w (a+b) = interference.w a \<union> interference.w b"  
    "interference.e (a+b) = interference.e a \<union> interference.e b"  
    by (auto split: interference.split simp: plus_interference_def)
  

  text \<open>Reported interference is consistent with observable state change. 
    Also includes that state change itself is consistent.\<close>
  definition "intf_consistent s i s' \<equiv> 
    let r = interference.r i in
    let w = interference.w i in
    let e = interference.e i in
      \<comment> \<open>Cells that are referred to must exist\<close>
      (\<forall>b \<in> addr.block ` (r\<union>w) \<union> e. block_state s' b \<noteq> block_state.FRESH) 
      \<comment> \<open>Cell state makes valid transitions only\<close>
    \<and> (\<forall>b. valid_block_trans s s' b)
      \<comment> \<open>Addresses not freed nor written must be unchanged\<close>
    \<and> (\<forall>a. is_valid_addr s a \<and> is_valid_addr s' a \<and> a\<notin>w \<longrightarrow> get_addr s a = get_addr s' a)
  "

  text \<open>Parallel state changes are feasible: the memory manager did not allocate the same blocks\<close>    
  definition "spar_feasible s s\<^sub>1 s\<^sub>2 \<equiv> alloc_blocks s s\<^sub>1 \<inter> alloc_blocks s s\<^sub>2 = {}"

  text \<open>Parallel executions have no race\<close>      
  definition "intf_norace s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2 \<equiv> 
    case i\<^sub>1 of (interference r\<^sub>1 w\<^sub>1 e\<^sub>1) \<Rightarrow> 
    case i\<^sub>2 of (interference r\<^sub>2 w\<^sub>2 e\<^sub>2) \<Rightarrow> 
    let fb\<^sub>1 = freed_blocks s s\<^sub>1 in
    let fb\<^sub>2 = freed_blocks s s\<^sub>2 in
    let f\<^sub>1 = {ADDR b i | b i. b\<in>fb\<^sub>1} in \<comment> \<open>Over-approx of addresses freed by thread1 \<close>
    let f\<^sub>2 = {ADDR b i | b i. b\<in>fb\<^sub>2} in \<comment> \<open>Over-approx of addresses freed by thread1 \<close>
      \<comment> \<open>Addresses read or modified by thread1 not modified by thread2\<close>
      (r\<^sub>1\<union>w\<^sub>1\<union>f\<^sub>1) \<inter> (w\<^sub>2\<union>f\<^sub>2) = {}
      \<comment> \<open>Addresses modified by thread1 not read or modified by thread2\<close>
    \<and> (w\<^sub>1\<union>f\<^sub>1) \<inter> (r\<^sub>2\<union>w\<^sub>2\<union>f\<^sub>2) = {}
      \<comment> \<open>Blocks referred to by thread1 not freed by thread2\<close>
    \<and> e\<^sub>1 \<inter> fb\<^sub>2 = {}
      \<comment> \<open>Blocks freed by thread1 not referred to by thread2\<close>
    \<and> fb\<^sub>1 \<inter> e\<^sub>2 = {}
    "
  
  definition "combine_states s\<^sub>1 i\<^sub>2 s\<^sub>2 \<equiv> mcombine s\<^sub>1 (interference.w i\<^sub>2) s\<^sub>2"

  lemma spar_feasible_sym: "spar_feasible s s\<^sub>1 s\<^sub>2 = spar_feasible s s\<^sub>2 s\<^sub>1"  
    by (auto simp: spar_feasible_def)

  lemma intf_norace_sym: "intf_norace s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2 = intf_norace s s\<^sub>2 i\<^sub>2 s\<^sub>1 i\<^sub>1"  
    by (simp add: intf_norace_def Let_def split: interference.split; fast)
    
    
  locale valid_parallel_execution =
    fixes s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2
    assumes consistent1: "intf_consistent s i\<^sub>1 s\<^sub>1"
    assumes consistent2: "intf_consistent s i\<^sub>2 s\<^sub>2"
    assumes feasible: "spar_feasible s s\<^sub>1 s\<^sub>2"
    assumes norace: "intf_norace s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2"
  begin  

    lemma valid_block_trans1: "valid_block_trans s s\<^sub>1 b"
      by (metis consistent1 intf_consistent_def)
  
    lemma valid_block_trans2: "valid_block_trans s s\<^sub>2 b"
      by (metis consistent2 intf_consistent_def)
  
    lemma par_blocks_exist_before:
      assumes [simp]: "is_ALLOC s\<^sub>1 b" 
      assumes [simp]: "is_ALLOC s\<^sub>2 b"  
      shows "is_ALLOC s b"
    proof (cases "block_state s b")
      case [simp]: FRESH
      
      have "b \<in> alloc_blocks s s\<^sub>1 \<inter> alloc_blocks s s\<^sub>2"
        unfolding alloc_blocks_def by auto
      with feasible have False
        unfolding spar_feasible_def by auto
      then show ?thesis ..
    next
      case FREED
      with valid_block_trans1[of b] have False unfolding valid_block_trans_def by simp
      then show ?thesis ..
    next
      case (ALLOC)
      then show ?thesis .
    qed
    
    lemma par_blocks_same_length:
      assumes "is_ALLOC s\<^sub>1 b" 
      assumes "is_ALLOC s\<^sub>2 b"  
      shows "block_size s\<^sub>1 b = block_size s\<^sub>2 b"  
      by (metis assms(1) assms(2) block_state.distinct(5) par_blocks_exist_before valid_block_trans1 valid_block_trans2 valid_block_trans_def)
        
    lemma write_disjoint: "interference.w i\<^sub>1 \<inter> interference.w i\<^sub>2 = {}"
      using norace unfolding intf_norace_def
      by (auto split: interference.splits simp: Let_def)
      
    lemma eq_nowrite1: "is_valid_addr s a \<Longrightarrow> is_valid_addr s\<^sub>1 a \<Longrightarrow> a\<notin>interference.w i\<^sub>1 \<Longrightarrow> get_addr s\<^sub>1 a = get_addr s a"  
      by (metis consistent1 intf_consistent_def)
      
    lemma eq_nowrite2: "is_valid_addr s a \<Longrightarrow> is_valid_addr s\<^sub>2 a \<Longrightarrow> a\<notin>interference.w i\<^sub>2 \<Longrightarrow> get_addr s\<^sub>2 a = get_addr s a"  
      by (metis consistent2 intf_consistent_def)

      
    lemma valid_mcombine[simp]: "is_valid_addr (mcombine s\<^sub>1 A s\<^sub>2) a \<longleftrightarrow> 
        (is_valid_addr s\<^sub>1 a \<and> \<not>is_FREED s\<^sub>2 (addr.block a)) 
      \<or> (is_valid_addr s\<^sub>2 a \<and> \<not>is_FREED s\<^sub>1 (addr.block a))"
      unfolding is_valid_addr_def
      by (cases a; auto split: block_state.split simp: par_blocks_same_length)
        
      
    lemma "is_valid_addr (mcombine s\<^sub>1 A s\<^sub>2) a \<Longrightarrow> get_addr (mcombine s\<^sub>1 A s\<^sub>2) a = (if a\<in>A then get_addr s\<^sub>2 a else get_addr s\<^sub>1 a)"    
      apply (cases a)
      subgoal for b i
        apply (cases "mapp s\<^sub>1 b"; cases "mapp s\<^sub>2 b"; simp)
        apply (auto simp: get_addr_def is_valid_addr_def block_state_def list_combine_nth par_blocks_same_length split: addr.split)
      
    
    
      
      
      
      
        
    lemma combine_states_sym: "combine_states s\<^sub>1 i\<^sub>2 s\<^sub>2 = combine_states s\<^sub>2 i\<^sub>1 s\<^sub>1"
      apply (clarsimp simp: mext) xxx, try to use block and addr characterization!
      subgoal for b
        unfolding combine_states_def
        apply simp
        apply (cases "mapp s\<^sub>1 b"; cases "mapp s\<^sub>2 b"; simp)
        proof goal_cases
          case [simp]: (1 vs\<^sub>1 vs\<^sub>2)
          
          (* TODO: we break the abstraction barrier of get_addr twice! *)
          
          obtain vs where [simp]: "mapp s b = BLOCK vs" using par_blocks_exist_before[OF 1] .
          
          have [simp]: "length vs\<^sub>1 = length vs" using cell_state_trans1[of b] by simp
          have [simp]: "length vs\<^sub>2 = length vs" using cell_state_trans2[of b] by simp
          
          have [simp]:
            "get_addr s\<^sub>1 (ADDR b i) = vs\<^sub>1!i" 
            "get_addr s\<^sub>2 (ADDR b i) = vs\<^sub>2!i" 
            for i by (simp_all add: get_addr_def)
          
          have [simp]:
            "is_valid_addr s (ADDR b i) \<longleftrightarrow> i<length vs" 
            "is_valid_addr s\<^sub>1 (ADDR b i) \<longleftrightarrow> i<length vs" 
            "is_valid_addr s\<^sub>2 (ADDR b i) \<longleftrightarrow> i<length vs" 
            for i   
            by (simp_all add: is_valid_addr_def)
            
          
          show ?case
            using write_disjoint eq_nowrite1[of "ADDR b _"]
            using write_disjoint eq_nowrite2[of "ADDR b _"]
            by (auto simp: list_eq_iff_nth_eq list_combine_nth)
          
        qed
      done
      
  end      
  
      
    
  (*  
      
      
    
  definition "valid_addr m a \<longleftrightarrow> cell.is_valid (mapp m a)"
  
  definition "m_fresh m \<equiv> {a. mapp m a = FRESH }"
  definition "m_freed m \<equiv> {a. mapp m a = FREED }"
  definition "m_valid m \<equiv> {a. cell.is_valid (mapp m a) }"
    
  lemma m_fresh_infinite[simp, intro!]: "infinite (m_fresh m)"
    unfolding m_fresh_def
    apply transfer
    using finite_Collect_not infinite_UNIV by blast
    
    
  datatype ('a,'idx) interference = interference (r: "('a\<times>'idx) set") (w: "('a\<times>'idx) set") (a: "'a set") (f: "'a set")
  hide_const (open) r w a f  
    
  abbreviation "intf_r r \<equiv> interference r {} {} {}"
  abbreviation "intf_w w \<equiv> interference {} w {} {}"
  abbreviation "intf_a a \<equiv> interference {} {} a {}"
  abbreviation "intf_f f \<equiv> interference {} {} {} f"

  instantiation interference :: (type,type) comm_monoid_add
  begin
    definition "0 = interference {} {} {} {}"
    definition "a+b \<equiv> case (a,b) of (interference ra wa aa fa, interference rb wb ab fb) \<Rightarrow>
      interference (ra\<union>rb) (wa\<union>wb) (aa\<union>ab) (fa\<union>fb)"

     instance
       apply standard     
       unfolding zero_interference_def plus_interference_def
       by (auto split: interference.split)
      
  end
  
    
  lemma interference_zero_simps[simp]:  
    "interference.a 0 = {}"  
    "interference.f 0 = {}"  
    "interference.r 0 = {}"  
    "interference.w 0 = {}"  
    by (auto split: interference.split simp: zero_interference_def)
    
  lemma interference_plus_simps[simp]:
    "interference.a (a+b) = interference.a a \<union> interference.a b"  
    "interference.f (a+b) = interference.f a \<union> interference.f b"  
    "interference.r (a+b) = interference.r a \<union> interference.r b"  
    "interference.w (a+b) = interference.w a \<union> interference.w b"  
    by (auto split: interference.split simp: plus_interference_def)
    
  
  text \<open>Over-approximation of everything that was modified: precise written addresses, and everything of allocated or freed blocks\<close>
  definition "intf_modified i \<equiv> case i of (interference rd wr al fr) \<Rightarrow> wr \<union> (al\<union>fr)\<times>UNIV"
    
  definition "intf_compatible i\<^sub>1 i\<^sub>2 \<equiv> interference.a i\<^sub>1 \<inter> interference.a i\<^sub>2 = {}"
  
  definition "intf_nointer i\<^sub>1 i\<^sub>2 \<equiv> 
    let m\<^sub>1 = intf_modified i\<^sub>1 in
    let m\<^sub>2 = intf_modified i\<^sub>2 in
    let r\<^sub>1 = interference.r i\<^sub>1 in
    let r\<^sub>2 = interference.r i\<^sub>2 in
      (r\<^sub>1 \<union> m\<^sub>1) \<inter> m\<^sub>2 = {}
    \<and> m\<^sub>1 \<inter> (r\<^sub>2\<union>m\<^sub>2) = {}
  "
  
  (*
  definition "intf_nointer i\<^sub>1 i\<^sub>2 \<equiv> case (i\<^sub>1,i\<^sub>2) of (interference r\<^sub>1 w\<^sub>1 a\<^sub>1 f\<^sub>1, interference r\<^sub>2 w\<^sub>2 a\<^sub>2 f\<^sub>2) \<Rightarrow>
    let ww\<^sub>1 = w\<^sub>1 \<union> a\<^sub>1 \<union> f\<^sub>1 in
    let ww\<^sub>2 = w\<^sub>2 \<union> a\<^sub>2 \<union> f\<^sub>2 in
      (r\<^sub>1 \<union> ww\<^sub>1) \<inter> ww\<^sub>2 = {}
    \<and> ww\<^sub>1 \<inter> (r\<^sub>2\<union>ww\<^sub>2) = {}
  "
  *)
   
  definition "intf_consistent s i s' \<equiv> case i of (interference rd wr al fr) \<Rightarrow>
    (finite (al\<union>fr))  
  \<and> (\<forall>a\<in>-(wr\<union>al\<union>fr). mapp s a = mapp s' a)
  \<and> (\<forall>a\<in>fr. mapp s a \<noteq> FREED \<and> mapp s' a = FREED)
  \<and> (\<forall>a\<in>al. mapp s a = FRESH)
  \<and> (\<forall>a\<in>(rd\<union>wr)-fr. cell.is_valid (mapp s' a))  
  \<and> (\<forall>a\<in>(rd\<union>wr)-al. cell.is_valid (mapp s a))
  "  

  definition "combine_states s\<^sub>1 i\<^sub>2 s\<^sub>2 \<equiv> mcombine s\<^sub>1 (intf_modified i\<^sub>2) s\<^sub>2"

  lemma intf_compatible_sym: "intf_compatible i\<^sub>1 i\<^sub>2 = intf_compatible i\<^sub>2 i\<^sub>1"  
    by (auto simp: intf_compatible_def)

  lemma intf_nointer_sym: "intf_nointer i\<^sub>1 i\<^sub>2 = intf_nointer i\<^sub>2 i\<^sub>1"  
    by (auto simp: intf_nointer_def Let_def split: interference.split)
  
  lemma combine_states_sym: "intf_nointer i\<^sub>1 i\<^sub>2 \<Longrightarrow> intf_consistent s i\<^sub>1 s\<^sub>1 \<Longrightarrow> intf_consistent s i\<^sub>2 s\<^sub>2 \<Longrightarrow>
    combine_states s\<^sub>1 i\<^sub>2 s\<^sub>2 = combine_states s\<^sub>2 i\<^sub>1 s\<^sub>1"
    unfolding intf_nointer_def intf_consistent_def combine_states_def mext intf_modified_def
    apply (auto split!: interference.splits simp: Let_def)
    by (metis Compl_iff Un_iff boolean_algebra.de_Morgan_disj inter_compl_diff_conv)
  
  *)
  

end
